#include <time.h>
#include <sstream>
#include "SolverPRT.h"

using namespace std;


SolverPRT::SolverPRT(): solution(""), solutionLength(0), currentTile(1), logging(false)
{ }

SolverPRT::~SolverPRT()
{ }

string SolverPRT::Solve()
{
	// (1) pokud jde vubec resit
	if(!CanSolvePuzzle())
	{
		return "Can't solve this puzzle...";
	}

	// (2) vynulovat pripadne reseni a najit pozici nuly
	solution.clear();
	puzzle.CountZeroPosition();


	// (3) hlavni smycka - rozdel a panuj
	unsigned int n = puzzle.GetN();
	// time_t start,startMini,end;
	// time(&start);
	clock_t start, startMini, end;
	start = clock();

	while(n>3)
	{
		/*
		std::stringstream debugFile;
		debugFile << "C:\\Puzzles\\!debug" << n << ".sbp";
		puzzle.SaveToSbp(debugFile.str());
		*/

		SolveFirstRow1();					// prvni polovina 1. radku
		if(!isTileOk(n-1) || !isTileOk(n))
			SolveFirstRow2();				// druha polovina 1. radku
		SolveFirstCol1();					// prvni polovina 1. sloupce
		if(!isTileOk(2*n+1) || !isTileOk(n+1))
			SolveFirstCol2();					// druha polovina 1. sloupce
		if(puzzle.TrimBoard())
		{
			--n;
			// cout << --n << "...\n";
		}
		else
		{
			cout << "Error! Can't trim board\n ";
			puzzle.Print();
			return "Error! Puzzle hasn't been solved...";
		}
	}
	// (4) pokud uz je to male --> resit "brutalni silou"
	if(n==3)
	{
		// puzzle.SaveToSbp("C:\\Puzzles\\!debug3x3.sbp");
		// time(&startMini);
		startMini = clock();
		string miniSolution = miniPuzzleSearch(puzzle.tiles);
		solution.append(miniSolution);
	}


	end = clock(); 
	// time(&end);
	// double dif1 = difftime (end,start);
	// double dif2 = difftime (end,startMini);
	double dif1 =  (double)((end - start) * 10 / CLOCKS_PER_SEC) / 10;
	double dif2 =  (double)((end - startMini) * 10 / CLOCKS_PER_SEC) / 10;
	// cout << "\n" << (double)diff_c / 10;
	printf("\nSolved in %.1f seconds ( %.1f of it solving 3x3 puzzle)",dif1,dif2);

	// (5) (pripadne overit a) vratit reseni
	return solution;
}

bool SolverPRT::CanSolvePuzzle() const
{
	return CanSolvePuzzle(this->puzzle);
}

bool SolverPRT::CanSolvePuzzle(const Puzzle &puzzle) const
{
	return (puzzle.IsSolvable() && puzzle.GetN()>2);
}

bool SolverPRT::isTileOk(int tile) const
{
	if(puzzle.GetN()>0) 
	{
		return (puzzle.GetGoalPosition(tile) == puzzle.GetPosition(tile));
	}
	else
		return false;
}

void SolverPRT::SolveFirstRow1()
{
	int n = puzzle.GetN();
	int tilePosition = 0;
	int goalPosition = 0;
	int goalCol, goalRow = 0;
	int tileRow, tileCol = 0;
	int zeroRow, zeroCol = 0;
	string moves = "";
	
	puzzle.PositionToCoord(puzzle.GetZeroPosition(),zeroRow,zeroCol);
	// prvnich n-1 kamenu
	for(int tileNumber=1; tileNumber<=n-1; ++tileNumber, ++goalPosition)
	{
		if(isTileOk(tileNumber))			// vlastne by nemelo nikdy nastat, leda na zacatku
		{
			// ale pokud je to posledni dilek, tak se musi posunout:
			if(tileNumber==n-1 && !isTileOk(n))
				ApplyMoves("LDR");
			continue;
		}

		goalCol = goalPosition;
		goalRow = 0;
		tilePosition = puzzle.GetPosition(tileNumber);
		puzzle.PositionToCoord(tilePosition,tileRow,tileCol);

		// 1) dostat mezeru napravo od kamene
		// 1.1 ... pozice mezery --- UZ MAM, bud z inicializace, nebo z minule pozice
		/* if(tileNumber==1)		// nemuzu se spolehnout na to, kde je 0
			puzzle.PositionToCoord(puzzle.GetZeroPosition(),zeroRow,zeroCol);
		else
		{	 
			// zeroRow = 0;
			// zeroCol = tileNumber-1;
		
			// ... NE, JSOU TAM HODNOTY NASTAVENE Z MINULA
		}*/

		// 1.2 ... presun na spravne misto
		MoveZeroRightNextTo(zeroRow,zeroCol,tileRow,tileCol);

		// 2) dostat do spravneho radku, nebo sloupce
		bool rightSide = tileCol>=goalCol;

		// SPESL. PRIPAD: kdy predposledni dilek uz byl v 1. radku -> je nutne ho posunout dolu
		if( (tileNumber==n-1) && (tileRow==0))
		{
			ApplyMoves("LDR");		// z pozice, kde je ...[n-1][mezera] na prvnim radku
									// do pozice [n-1][mezera] ... v posl sloupci
			return;
		}
		
		// pokud pujdeme zleva, nastavime cilovy radek na 2. odshora, protoze jinak by
		// se mohlo pokazit to, co uz tam je postavene
		// a do 2. radku take v pripade, ze je to predpoledni kamen!
		if(!rightSide || tileNumber==n-1)
			goalRow++;

		// jit sikmo nahoru, dokud neni ve spravnem sloupci, nebo radku
		while( (tileRow!=goalRow) && (tileCol!=goalCol) )
		{
			// pokud to bylo v prave polovine
			if(rightSide) // jit doleva nahoru
			{
				ApplyMoves(MACRO_UL_R);
				--tileCol;
			}
			else		 // jit doprava nahoru
			{
				ApplyMoves(MACRO_UR_R);
				++tileCol;
			}
			--tileRow;
		}

		// 3) uz jen posun horizontalne / vertikalne
		moves.clear();
		if(tileRow==goalRow)
		{
			if(tileCol>goalCol)				// doleva
			{
				AppendMoves(moves,MACRO_L_R,tileCol-goalCol);		// 
			}
			else							// doprava 
			{
				AppendMoves(moves,MACRO_R_Rd,goalCol-tileCol);
				// ale jestli to bylo doprava, je jeste nutne posunout o 1 nahoru (krome pripadu n-1):
				if(tileNumber!=n-1 && !rightSide)
					moves.append(MACRO_UP_R);
			}
		}
		else // if(tileCol==goalCol)
		{
			// neni jina moznost, nez ze se musi posunout az nahoru (krome pripadu n-1)
			if(tileNumber!=n-1)
				AppendMoves(moves,MACRO_UP_R,tileRow-0);				//
			else
				AppendMoves(moves,MACRO_UP_R,tileRow-1);				//
		}
		ApplyMoves(moves);
		
		// pozice mezery:
		zeroRow = 0;
		zeroCol = tileNumber;
		
	}
	return;
}

void SolverPRT::SolveFirstRow2()
{
	// vyresi koncovku 1. rady:

	// predpoklad: v 2. radku je na konci [kamen n-1][mezera]
	//			NEBO posledni sloupec je [kamen n-1] a pod tim [mezera]

	int n = puzzle.GetN();
	bool config1 = (puzzle[2*n-1]==0) && (puzzle[2*n-2]==n-1);		// std. ocekavana konfigurace
	bool config2 = (puzzle[n-1]==n-1) && (puzzle[2*n-1]==0);		// pokud uz cislo n-1 bylo na "svem" miste
	if((!config1) && (!config2))		// neni ocekavana konfigurace
	{
		cout << "\nError in function SolveFirstRow2() (n=" << n << ") : unexpected board configuration...";
		return;
	}
	else
	{
		if(puzzle[n-1]==n)
		{
			// specialni pripad (soubor __koncovka(radek).sbp), osetri se makrem:
			ApplyMoves(MACRO_ROWEND_SPECIAL);
		}
		else if(config2 && puzzle[n-2]==n) // SPESL PRIPAD PRO CONFIG 2 (__koncovka(radek)2.sbp)
		{
			ApplyMoves(MACRO_ROWEND_SPECIAL2);
		}
		else
		{
			// * nejprve dostat predposledni kamen do praveho horniho rohu, pokud uz tam neni
			if(config1)
				ApplyMoves(MACRO_ROTATECC_DR);
			
			// * najit posledni kamen
			int tileRow, tileCol = 0;
			int goalRow = 1;		// = zeroRow
			int goalCol = n-1;		// = zeroCol
			currentTile = n;
			int tilePosition = puzzle.GetPosition(currentTile);
			puzzle.PositionToCoord(tilePosition,tileRow,tileCol);

			// * dostat ho pod ten predposledni, tj. na pozici (1,n-1)
			if(tileCol==n-1)
			{
				LiftLastTile(tileRow,goalRow);
			}
			else
			{
				MoveZeroRightNextTo(goalRow,goalCol,tileRow,tileCol);
				// jit sikmo nahoru doprava, dokud neni ve spravnem sloupci, nebo radku
				// z tech. duvodu nutne nastavit cil, sloupec na n-2:
				goalCol=n-2;	// pze do toho posl. nelze jit makrem!
				while( (tileRow!=goalRow) && (tileCol!=goalCol) )
				{
					ApplyMoves(MACRO_UR_R);
					++tileCol;
					--tileRow;
				}
				if(tileCol==n-2)
				{
					ApplyMoves("L");					// tim se kamen dostane do posl. sloupce
					for(; tileRow>goalRow; --tileRow)		// posun nahoru
					{
						ApplyMoves(MACRO_UP_L);
					}
				}
				else	// if tileRow==goalRow (==1)
				{
					ShiftTile(tileCol,goalCol);			// pze do posledniho sloupce to nejde
					ApplyMoves("L");
				}
			}

			// * provest zaverecnou "rotaci"
			// predpoklad: mezera je na pozici vlevo od posl. kamene
			if(puzzle[2*n-2]==0)
			{
				ApplyMoves(MACRO_ROWEND_STD);
			}
			else
			{
				cout << "\nError in function SolveFirstRow2() (n=" << n << ") : can't do final rotation...";
			}
		}
	}
	return;
}

void SolverPRT::SolveFirstCol1()
{
	int n = puzzle.GetN();
	int tilePosition = 0;
	int goalPosition = 0;
	int goalCol, goalRow = 0;
	int tileRow, tileCol = 0;
	int zeroRow, zeroCol = 0;
	string moves = "";

	puzzle.PositionToCoord(puzzle.GetZeroPosition(),zeroRow,zeroCol);
	
	// prvnich n/2 kamenu
	for(int i=1; i<=n-2; ++i)							// -1 za 1. uz vyplneny radek a -1 za horni dilek
	{
		int tileNumber = (n*n) - (i*n) + 1;				// pro n=4: 13 ... 9 ... ale 5 uz ne
		if(isTileOk(tileNumber))
		{
			if((tileNumber==2*n+1) && !isTileOk(n+1))
			{
				tileRow = 2, tileCol = 0;
				MoveZeroUpNextTo(zeroRow,zeroCol,tileRow,tileCol);
				ApplyMoves("DRU");						// aby se to dostalo do pozadovane konfigurace
			}
			continue;
		}

		goalCol = 0;
		goalRow = tileNumber/n;
		tilePosition = puzzle.GetPosition(tileNumber);
		puzzle.PositionToCoord(tilePosition,tileRow,tileCol);

		// 1) dostat mezeru nahoru od kamene
		// 1.1 ... pozice mezery -- UZ MAM
		
		// 1.2 ... presun na spravne misto
		MoveZeroUpNextTo(zeroRow,zeroCol,tileRow,tileCol);

		// 2) dostat do spravneho radku, nebo sloupce
		bool topSide = tileRow<=goalRow;

		// SPESL. PRIPAD: kdy predposledni dilek uz byl v 1. radku -> je nutne ho posunout dolu
		if( (tileNumber==2*n+1) && (tileCol==0))
		{
			ApplyMoves("DRU");	// obdobne, jako u FirstRow1()
			return;
		}
		
		// pokud pujdeme zleva, nastavime cilovy radek na 2. odshora, protoze jinak by
		// se mohlo pokazit to, co uz tam je postavene
		if(!topSide || tileNumber==2*n+1)
			goalCol++;

		// jit sikmo doleva, dokud neni ve spravnem sloupci, nebo radku
		while( (tileRow!=goalRow) && (tileCol!=goalCol) )
		{
			// pokud to bylo v prave polovine
			if(topSide) // jit doleva dolu
			{
				ApplyMoves(MACRO_DL_U);
				++tileRow;
			}
			else		 // jit doprava nahoru
			{
				ApplyMoves(MACRO_UL_U);
				--tileRow;
			}
			--tileCol;
		}

		// 3) uz jen posun horizontalne / vertikalne
		moves.clear();
		if(tileCol==goalCol)
		{
			if(tileRow<goalRow)			// dolu
			{
				AppendMoves(moves,MACRO_D_U,goalRow-tileRow);		// 
			}
			else							// nahoru 
			{
				AppendMoves(moves,MACRO_U_U,tileRow-goalRow);
				// ale pze to bylo nahoru, je jeste nutne posunout o 1 doleva: (pac jsme ve 2.sloupci)
				if(tileNumber!=2*n+1 && !topSide)
					moves.append(MACRO_L_U);		// ale ten predposledni tam chci nechat!
			}
		}
		else // if(tileRow==goalRow)
		{
			// neni jina moznost, nez ze se musi posunout az uplne doleva
			if(tileNumber==2*n+1)
			{
				// krome pripadu predposledniho kamene
				if(tileCol>=1)
				{
					AppendMoves(moves,MACRO_L_U,tileCol-1);
				}
				else	// kdyz byl v 1. sloupci -> nutno osetrit zvlast
				{

				}
			}
			else
			{
				AppendMoves(moves,MACRO_L_U,tileCol-0);					//
			}
		}
		ApplyMoves(moves);
		
		// pozice mezery:
		zeroRow = n - i - 1;					// pze pozice se pocitaji od 0
		zeroCol = 0;
		
	}
	return;
}

void SolverPRT::SolveFirstCol2()
{
	// vyresi koncovku 1. sloupce:

	// predpoklad: v 2. sloupci je nahore [mezera][kamen+1]
	//			NEBO prvni volny (tzn. 2.) radek je [kamen+1][mezera]
	
	int n = puzzle.GetN();
	bool config1 = (puzzle[n+1]==0) && (puzzle[2*n+1]==2*n+1);		// std. ocekavana konfigurace
	bool config2 = (puzzle[n]==2*n+1) && (puzzle[n+1]==0);			// pokud uz cislo 2*n+1 bylo na "svem" miste
	if((!config1) && (!config2))		// neni ocekavana konfigurace
	{
		cout << "\nError in function SolveFirstCol2() (n=" << n << ") : unexpected board configuration...";
		return;
	}
	else
	{
		if(puzzle[n]==n+1)
		{
			// specialni pripad, osetri se makrem:
			ApplyMoves(MACRO_COLEND_SPECIAL);
		}
		else if(config2 && puzzle[2*n]==n+1) // SPESL PRIPAD PRO CONFIG 2	()
		{
			ApplyMoves(MACRO_COLEND_SPECIAL2);
		}
		else
		{
			// * nejprve dostat predposledni kamen do leveho horniho rohu, pokud uz tam neni
			if(config1)
				ApplyMoves(MACRO_ROTATECC_UR);
			
			// * najit posledni kamen
			int tileRow, tileCol = 0;
			int goalRow = 1;		// = zeroRow
			int goalCol = 1;		// = zeroCol
			currentTile = n+1;
			int tilePosition = puzzle.GetPosition(currentTile);
			puzzle.PositionToCoord(tilePosition,tileRow,tileCol);

			// * dostat ho na pozici (2,2)
			if(tileRow==1)
			{
				MoveZeroRightNextTo(goalRow,goalCol,tileRow,tileCol);
				ShiftTile(tileCol,goalCol);
				ApplyMoves("DL");			// to abych dostal mezeru pod kamen n+1
			}
			else
			{
				MoveZeroUpNextTo(goalRow,goalCol,tileRow,tileCol);
				// jit sikmo nahoru doprava, dokud neni ve spravnem sloupci, nebo radku
				// z tech. duvodu nutne nastavit cil, sloupec na n-2:
				goalRow=2;		// pze do prvniho volneho radku nelze jit makrem
				while( (tileRow!=goalRow) && (tileCol!=goalCol) )
				{
					ApplyMoves(MACRO_UL_U);
					--tileCol;
					--tileRow;
				}
				if(tileRow==2)
				{
					ApplyMoves("D");					// tim se kamen dostane do druheho radku
					for(;tileCol>1;--tileCol)
					{
						ApplyMoves(MACRO_L_D);
					}
				}
				else	// if tileCol==goalCol (==1)
				{
					ApplyMoves("D");					// tim se dostanu pod ten kamen
					--tileRow;
					for(;tileRow>1;--tileRow)			// posun nahoru, jak to jde
					{
						ApplyMoves(MACRO_U_D);
					}
				}
			}

			// * provest zaverecnou "rotaci"
			// predpoklad: mezera je na pozici dole od 1. kamene ve sloupci
			if(puzzle[2*n+1]==0)
			{
				ApplyMoves(MACRO_COLEND_STD);
			}
			else
			{
				cout << "\nError in function SolveFirstRow2() (n=" << n << ") : can't do final rotation...";
			}
		}
	}



}

void SolverPRT::ApplyMoves(std::string moves)
{
	if(logging)
	{
		loggedMoves.append(moves);
	}
	if(puzzle.ApplySequence(moves))
	{
		solution.append(moves);
	}
}
void SolverPRT::MoveZeroUpNextTo(int zeroRow, int zeroCol, int& row, int& col)
{
	string moves = "";
	// 1) nejprve do stejneho sloupce:
	
	// ale blby, kdyz uz jsou ve stejnem radku:
	if(zeroRow==row)
	{
		if(zeroRow==1)					// to je 2. radek od shora - musim nejprve krok dolu
		{
			ApplyMoves("D");
			zeroRow++;
		}
		else
		{
			ApplyMoves("U");
			zeroRow--;
		}
	}
	
	// nevic bacha, aby nezboril, co uz mam v 1. sloupci postavene:
	bool firstColHack = (zeroRow>row) && zeroCol!=0 && (col==0);
	//		tedy ve chvili, kdy mezera je "vpravo" od uz postaveneho useku a cilovy
	//		kamen se nachazi v 1. sloupci

	// jakym smerem to bude:
	char move = (zeroCol>col) ? 'L' : 'R';
	// nemuzu jit az nahoru, kdyz bych si znicil, co uz tam je postavene
	int kroku = firstColHack ? abs(zeroCol-col)-1 : abs(zeroCol-col);
	moves.append(kroku,move);
	
	// 2.a) nahoru
	if(zeroRow>=row)
	{
		if(firstColHack)
		{
			moves.append(zeroRow-row-1,'U');		// tim se dostanu sikmo vpravo dole od kamene
			moves.append("LU");
		}
		else
		{
			moves.append(zeroRow-row,'U');
		}
		// tim padem se ale o 1 posunul ten kamen:
		++row;
	}
	// 2.b) dolu
	else
	{
		moves.append(row-zeroRow-1,'D');
	}

	ApplyMoves(moves);
}

void SolverPRT::MoveZeroRightNextTo(int zeroRow, int zeroCol, int& row, int& col)
{
	string moves = "";
	// 1) nejprve do stejneho radku:
	
	// ale blby, kdyz uz jsou ve stejnem sloupci:
	if(zeroCol==col)
	{
		if((zeroCol==puzzle.GetN()-1) && (col==puzzle.GetN()-1))
		{
			ApplyMoves("L");
			zeroCol--;
		}
		else
		{
			ApplyMoves("R");
			zeroCol++;
		}
	}

	// navic bacha, aby nezboril, co uz mam na 1. radku postavene:
	bool firstRowHack = (zeroCol<col-1) && zeroRow!=0 && (row==0);
	//		tedy ve chvili, kdy mezera je "pod" uz postavenym usekem a cilovy
	//		kamen se nachazi v 1. radku
	
	// jakym smerem to bude:
	char move = (zeroRow>row) ? 'U' : 'D';
	// nemuzu jit az nahoru, kdyz bych si znicil, co uz tam je postavene
	int kroku = firstRowHack ? abs(zeroRow-row)-1 : abs(zeroRow-row);
	moves.append(kroku,move);
	
	// 2.a) doprava
	if(zeroCol<col)
	{
		if(firstRowHack)
		{
			moves.append(col-zeroCol-1,'R');		// tim se dostanu sikmo vlevo od kamene
			moves.append("UR");
		}
		else
		{
			// normalka jedem doprava
			moves.append(col-zeroCol,'R');
		}
		// tim padem se ale o 1 posunul ten kamen:
		--col;
	}
	// 2.b) doleva
	else
	{
		moves.append(zeroCol-col-1,'L');
	}

	ApplyMoves(moves);
}


void SolverPRT::MoveZeroLeftNextTo(int zeroRow, int zeroCol, int& row, int& col)
{
	string moves = "";
	// 1) nejprve do stejneho radku:
	
	// ale blby, kdyz uz jsou ve stejnem sloupci:
	if(zeroCol==col)
	{
		if((zeroCol==puzzle.GetN()-1) && (col==puzzle.GetN()-1))
		{
			ApplyMoves("L");
			zeroCol--;
		}
		else
		{
			ApplyMoves("R");
			zeroCol++;
		}
	}

	// navic bacha, aby nezboril, co uz mam na 1. radku postavene:
	bool firstRowHack = (zeroCol<col-1) && zeroRow!=0 && (row==0);

	
	// jakym smerem to bude:
	char move = (zeroRow>row) ? 'U' : 'D';
	// nemuzu jit az nahoru, kdyz bych si znicil, co uz tam je postavene
	int kroku = firstRowHack ? abs(zeroRow-row)-1 : abs(zeroRow-row);
	moves.append(kroku,move);
	
	// 2.a) doprava
	if(zeroCol<col)
	{
		if(firstRowHack)
		{
			moves.append(col-zeroCol-1,'R');		// tim se dostanu sikmo vlevo od kamene
			moves.append("U");
		}
		else
		{
			// normalka jedem doprava
			moves.append(col-zeroCol-1,'R');
		}
	}
	// 2.b) doleva
	else
	{
		moves.append(zeroCol-col,'L');
		// tim padem se ale o 1 posunul ten kamen:
		++col;
	}

	ApplyMoves(moves);
}

void SolverPRT::LiftLastTile(int tileRow, int goalRow)
{
	// fce "vytahne" posledni kamen (n) v poslednim sloupci z radku tileRow do radku goalRow

	// predpoklada, ze na pozici (1,n-1) je mezera!
	if(puzzle[2*puzzle.GetN()-1]==0)
	{
		string moves = "";
		AppendMoves(moves,"D",tileRow-goalRow);
		tileRow--;
		moves.append("LU");		// abych se dostal nalevo od kamene
		for(;tileRow>goalRow;--tileRow)
		{
			moves.append(MACRO_UP_L);
		}
		ApplyMoves(moves);
	}
}

void SolverPRT::ShiftTile(int tileCol, int goalCol)
{
	// posune horizontalne dany kamen do ciloveho sloupce
	// predpoklad: mezera je umistena napravo od kamene!

	// !!!!! nelze pouzit pro posledni radek !!!!!!
	// !!!!! nelze pouzit pro presun z/do posl. sloupce !!!!!
	string moves = "";
	string macroMove = (tileCol>=goalCol) ? MACRO_L_R : MACRO_R_Rd;
	for(int i=0; i<abs(tileCol-goalCol);++i)
	{
		moves.append(macroMove);
	}
	ApplyMoves(moves);
}


void AppendMoves(string& solution, const string& moves, int times)
{
	if(times<1)
			return;
	// jen pomocna fuknce, ktera za retezec [solution] prida [times] kopii retezce [moves]
	for(int i=0; i<times; ++i)
	{
		solution.append(moves);
	}
}

void SolverPRT::StartLog()
{
	savedPuzzle = puzzle;
	savedSolution = solution;
	logging = true;
	loggedMoves.clear();
}


void SolverPRT::StopLog()
{
	logging = false;
}


int SolverPRT::GetLogLength() const
{
	return loggedMoves.length();
}


int SolverPRT::GetLogLengthOptimized() const
{
	return optimizeSolution(loggedMoves).length();
}

void SolverPRT::RestoreSavedPoint()
{
	puzzle = savedPuzzle;
	solution = savedSolution;
}