#ifndef SOLVERPRT_H
#define SOLVERPRT_H

#include "puzzle.h"
#include "miniPuzzle.h"


class SolverPRT
{
public:
	SolverPRT();
	~SolverPRT();

	std::string Solve();

	std::string GetSolution() const { return this->solution; }
	void SetPuzzle(const Puzzle& _puzzle) { this->puzzle = _puzzle; }
	int GetSolutionLength() const { return this->solutionLength; }
	bool CanSolvePuzzle() const;
	bool CanSolvePuzzle(const Puzzle& puzzle) const;
	bool isTileOk(int tile) const;


protected:

	void SolveFirstRow1();
	void SolveFirstRow2();
	void SolveFirstCol1();
	void SolveFirstCol2();
	

	// pohyby
	void ApplyMoves(std::string moves);

	// jednotlive subsekvence:
	void MoveZeroRightNextTo(int zeroRow, int zeroCol, int& row, int& col);
	void MoveZeroLeftNextTo(int zeroRow, int zeroCol, int& row, int& col);
	void MoveZeroUpNextTo(int zeroRow, int zeroCol, int& row, int& col);
	void LiftLastTile(int tileRow, int goalRow);
	void ShiftTile(int tileCol, int goalCol);


	// pro vnitrni logovani a moznost vraceni se
	void StartLog();
	void StopLog();
	int GetLogLength() const;
	int GetLogLengthOptimized() const;
	void RestoreSavedPoint();

	string loggedMoves;				// zaznam tahu
	bool logging;					// priznak, jestli se loguje, nebo ne
	Puzzle savedPuzzle;				// ulozena konfigurace
	std::string savedSolution;		// ulozena cast reseni
	
	Puzzle puzzle;
	std::string solution;
	int solutionLength;

	int currentTile;

};

void AppendMoves(string& solution, const string& moves, int times);


// ********************************************************************
// ****                    konstantni makra:                       ****
// ********************************************************************

const string MACRO_UP_R = "ULDRU";						// posun UP, mezera je R od kamene
const string MACRO_UP_L = "URDLU";						// posun UP, mezera je L od kamene
const string MACRO_L_R = "DLLUR";						// posun LEFT, mezera je R od kamene
const string MACRO_L_Rd = "DLLUR";						// posun LEFT, mezera je R od kamene a presouva se spodem
const string MACRO_L_Ru = "ULLDR";						// posun LEFT, mezera je R od kamene a presouva se horem
const string MACRO_L_Ld = "RDLLU";						// posun LEFT, mezera je L od kamene a presouva se dolem
const string MACRO_L_Lu = "RULLD";						// posun LEFT, mezera je L od kamene a presouva se horem

const string MACRO_R_Rd = "LDRRU";						// posun RIGHT, mezera je R od kamene a presouva se spodem
const string MACRO_R_Ru = "LURRD";						// posun RIGHT, mezera je R od kamene a presouva se horem
const string MACRO_R_Ld = "DRRUL";						// posun RIGHT, mezera je L od kamene a presouva se spodem
const string MACRO_R_Lu = "URRDL";						// posun RIGHT, mezera je L od kamene a presouva se horem

const string MACRO_D_L = "DRULD";						// posun DOWN, mezera je L od kamene
const string MACRO_D_R = "DLURD";						// posun DOWN, mezera je R od kamene
const string MACRO_U_L = "URDLU";						// posun UP, mezera je L od kamene
const string MACRO_U_R = "ULDRU";						// posun UP, mezera je R od kamene



const string MACRO_UR_U = "DRULUR";						// posun UP+R, mezera je R od kamene
const string MACRO_UR_R = "LURDRU";						// posun UP+R, mezera je R od kamene
const string MACRO_UR_L = "URDRUL";						// posun UP+R, mezera je L od kamene
const string MACRO_UL_U = "DLURUL";						// posun UP+L, mezera je U od kamene
const string MACRO_UL_R = "ULDLUR";						// posun UP+L, mezera je R od kamene
const string MACRO_UL_L = "RULDLU";						// posun UP+L, mezera je L od kamene
const string MACRO_DL_U = "LDRDLU";						// posun DOWN+L, mezera je U od kamene
const string MACRO_DL_L = "RDLULD";						// posun DOWN+L, mezera je L od kamene
const string MACRO_DL_R = "DLULDR";						// posun DOWN+L, mezera je R od kamene
const string MACRO_DR_U = "RDLDRU";						// posun UP+L, mezera je U od kamene
const string MACRO_DR_R = "LDRURD";						// posun UP+L, mezera je R od kamene
const string MACRO_DR_L = "DRURDL";						// posun UP+L, mezera je L od kamene


const string MACRO_D_U = "RDDLU";						// posun DOWN, mezera je U od kamene
const string MACRO_U_U = "DRUUL";						// posun UP, mezera je U od kamene
const string MACRO_L_U = "LDRUL";						// posun LEFT, mezera je U od kamene
const string MACRO_U_D = "RUULD";						// pusun UP, mezera je D od kamene
const string MACRO_U_Dl = "LUURD";						// pusun UP, mezera je D od kamene a presouva se po leve strane
const string MACRO_L_D = "LURDL";						// posun LEFT, mezera je D od kamene
const string MACRO_ROWEND_SPECIAL = "ULDLURRDLULD";		// koncovka pro 1. radu, specialni pripad
const string MACRO_ROWEND_SPECIAL2 = "LULDDRRULLURRD";  // koncovka pro 1. radu, special. pripad 2 (config2)
const string MACRO_ROWEND_STD = "URD";					// standardni koncova rotace pro 1. radek
const string MACRO_COLEND_SPECIAL = "RDLLDRURDLLUR";	// koncovka pro 1. sloupec, specialni pripad
const string MACRO_COLEND_SPECIAL2 = "DLDRRUULDDLUUR";	// koncovka pro 1. sloupec, specialni pripad
const string MACRO_COLEND_STD = "LUR";					// standardni koncova rotace pro 1. sloupec

const string MACRO_ROTATECC_DR = "LURD";				// rotace proti smeru, mezera DOWN RIGHT
const string MACRO_ROTATECC_UR = "DLUR";				// rotace proti smeru, mezera UP RIGHT
const string MACRO_ROTATEC_UR =  "LDRU";				// rotace po smeru, mezera UP RIGHT
const string MACRO_ROTATEC_UL =  "DRUL";				// rotace po smeru, mezera UP LEFT
const string MACRO_ROTATEC_DR =  "ULDR";				// rotace po smeru, mezera DOWN RIGHT
const string MACRO_ROTATEC_DL =  "RULD";				// rotace po smeru, mezera DOWN LEFT

#endif