#include "miniPuzzle.h"


void PositionToCoord(int pos, int &row, int &col)
{
	col = pos % MINIBOARD_N;
	row = (pos / MINIBOARD_N);
	return;
}

GraphNode * newGraphNode(const GraphNode* _node, const char move)
{
	if(_node->CanUseMove(move))
	{
		char lastMove = 'X';
		if(_node->path.length()>0)
			lastMove=_node->path[_node->path.length()-1];

		// neopakovat predchozi tah (resp. nevracet se):
		if( (move=='U' && lastMove=='D') || (move=='D' && lastMove=='U') ||
			(move=='L' && lastMove=='R') || (move=='R' && lastMove=='L') )
		{
			return 0;	// nevracet se
		}
		else
		{
			GraphNode* newNode = new GraphNode(*_node);

			if(newNode->DoMove(move))
			{
				newNode->path.push_back(move);
				newNode->g +=1;
				newNode->h = newNode->heur();
				return newNode;
			}
			else
			{
				delete newNode;
				return 0;
				cout << "To nejde...";
			}
		}
	}
	return 0;
}


GraphNode::GraphNode(): g(0), zero(0), h(-1)
{ };

GraphNode::GraphNode(int _g, string _path): g(_g), path(_path), zero(0), h(-1)
{ };

GraphNode::~GraphNode()
{ };



bool GraphNode::CanUseMove(const char move) const
{
	if(move=='U')
	{
		return (zero>=MINIBOARD_N);
	}
	else if(move=='L')
	{
		return (zero%MINIBOARD_N!=0);
	}
	else if(move=='D')
	{
		return (zero<MINIBOARD_SIZE-MINIBOARD_N);
	}
	else if(move=='R')
	{
		return (zero%MINIBOARD_N!=MINIBOARD_N-1);
	}
	else
		return false;
}

bool GraphNode::DoMove(const char move)
{
	if(CanUseMove(move))
	{
		switch(move)
		{
			case 'U':
						tiles[zero]=tiles[zero-MINIBOARD_N];
						tiles[zero-MINIBOARD_N]=0;
						zero=zero-MINIBOARD_N;
						return true;
						break;					
			case 'D':
						tiles[zero]=tiles[zero+MINIBOARD_N];
						tiles[zero+MINIBOARD_N]=0;
						zero=zero+MINIBOARD_N;
						return true;
						break;
			case 'L':
						tiles[zero]=tiles[zero-1];
						tiles[zero-1]=0;
						zero--;
						return true;
						break;
			case 'R':
						tiles[zero]=tiles[zero+1];
						tiles[zero+1]=0;
						zero++;
						return true;
						break;
			default:	return false;

		}
	}
	else
		return false;
}

unsigned GraphNode::heur() const
{
	unsigned int dist=0;
	unsigned int goalPosition = 0;
	int radek_1, radek_2, sloupec_1, sloupec_2 = 0;
	for(unsigned i=0; i<MINIBOARD_SIZE;++i)
	{
		if(tiles[i]==0)
				continue;


		goalPosition = ((tiles[i]-1) % MINIBOARD_N) + MINIBOARD_N * ((tiles[i]-1) / MINIBOARD_N);
		PositionToCoord(i,radek_1,sloupec_1);
		PositionToCoord(goalPosition,radek_2,sloupec_2);
		dist += abs(radek_1-radek_2) + abs(sloupec_1-sloupec_2);
	}

	// h=dist;
	return dist;
}

int GraphNode::f() const
{
	int dist = h>=0 ? h : heur();
	return dist + g;
}


// ********************************************************************************
//  SAMOTNY ALGORITMUS A*
// ********************************************************************************

std::string miniPuzzleSearch(const vector<int>& board)
{
	char moves[4] = {'U','D','L','R'};
	// *** (1) udelat z toho pocatecni stav ***
	GraphNode* initialNode = new GraphNode(0,"");

	// narvat tam pocatecni konfiguraci:
	for(unsigned i=0; i<board.size() && i<MINIBOARD_SIZE; ++i)
	{
		initialNode->tiles[i] = board[i];
		if(board[i]==0)
			initialNode->zero=i;
	}

	// spocitat heuristiku:
	initialNode->h = initialNode->heur();

	NODEQUEUE openList;
	openList.push(initialNode);

	GraphNode* currentNode = 0;
	while(!openList.empty())
	{
		// (1) vzit prvni vrchol na rade
		currentNode=openList.top();
		openList.pop();
		
		// (2) test, jestli nahodou nejsme v cili:
		if(currentNode->h==0)
		{
			// JUPIII!
			string solution = currentNode->path;
			
			//uklidit po sobe:
			delete currentNode;
			while(!openList.empty())
			{
				currentNode=openList.top();
				openList.pop();
				delete currentNode;
			}
			
			return solution;
		}

		// (3) zkusit vsechny mozne tahy:
		GraphNode* newNode = 0;
		for(int i=0; i<4; ++i)
		{
			if(newNode = newGraphNode(currentNode,moves[i]))
			{
				openList.push(newNode);
			}
		}

		// a smazat tento vrchol:
		delete currentNode;
	}

	return "";

}
